#!/usr/bin/env python
# coding: utf-8

import numpy as np
from statsmodels.api import OLS, add_constant
from statsmodels.tsa.stattools import adfuller

from napoleontoolbox.cointegration_utils import cointegration_utilities
from napoleontoolbox.signal import signal_utility

from npxlogger import log


def estimate_long_run_short_run_relationships(x, y):

    try:
        assert isinstance(y, np.ndarray)
    except:
        log.error('Input series y should be of type np.ndarray')

    try:
        assert isinstance(x, np.ndarray)
    except:
        log.error('Input series x should be of type np.ndarray')

    try:
        assert np.isnan(np.sum(y)) == False
    except:
        log.error('Input series y has nan-values. Unhandled case.')

    try:
        assert np.isnan(np.sum(y)) == False
    except:
        log.error('Input series x has nan-values. Unhandled case.')

    try:
        assert len(y)==len(x)
    except:
        log.error('The two input series y and x do not have the same legnth.')

    long_run_ols = OLS(y, add_constant(x), has_const=True)
    long_run_ols_fit = long_run_ols.fit()

    c, gamma = long_run_ols_fit.params
    z = long_run_ols_fit.resid

    short_run_ols = OLS(np.diff(y), (z[:-1]))
    short_run_ols_fit = short_run_ols.fit()

    alpha = short_run_ols_fit.params[0]

    return c, gamma, alpha, z



def engle_granger_two_step_cointegration_test(x, y):

    try:
        assert isinstance(y, np.ndarray)
    except:
        log.error('Input series y should be of type np.ndarray')

    try:
        assert isinstance(x, np.ndarray)
    except:
        log.error('Input series x should be of type np.ndarray')

    try:
        assert np.isnan(np.sum(y)) == False
    except:
        log.error('Input series y has nan-values. Unhandled case.')

    try:
        assert np.isnan(np.sum(y)) == False
    except:
        log.error('Input series x has nan-values. Unhandled case.')

    try:
        assert len(y) == len(x)
    except:
        log.error('The two input series y and x do not have the same legnth.')

    c, gamma, alpha, z = estimate_long_run_short_run_relationships(x, y)

    # NOTE: The p-value returned by the adfuller function assumes we do not estimate z first, but test
    # stationarity of an unestimated series directly. This assumption should have limited effect for high N,
    # so for the purposes of this course this p-value can be used for the EG-test. Critical values taking
    # this into account more accurately are provided in e.g. McKinnon (1990) and Engle & Yoo (1987).

    adfstat, pvalue, usedlag, nobs, crit_values, icbest = adfuller(z)

    return adfstat, pvalue, crit_values, alpha, z

def cointegration_test(x, y, risk = 10, alpha = 0.05):

    adfstat, pvalue, crit_values, alpha_ols, z = engle_granger_two_step_cointegration_test(x,y)

    if pvalue > alpha:
        return False, alpha_ols, z
    elif adfstat > crit_values[str(risk)+'%']:
        return False, alpha_ols, z
    else:
        return True, alpha_ols, z

def simple_return(x_close, y_close):
    y_return = np.concatenate([np.array([np.NaN]),np.diff(y_close) / y_close[:-1]])
    x_return = np.concatenate([np.array([np.NaN]),np.diff(x_close) / x_close[:-1]])
    y_return = np.nan_to_num(y_return)
    x_return = np.nan_to_num(x_return)
    return x_return, y_return

def log_return(x_close, y_close):
    y_log_return = np.concatenate([np.array([np.NaN]),np.diff(np.log(y_close))])
    x_log_return = np.concatenate([np.array([np.NaN]),np.diff(np.log(x_close))])
    y_log_return = np.nan_to_num(y_log_return)
    x_log_return = np.nan_to_num(x_log_return)
    return x_log_return, y_log_return

def log_return_regressor(x_log_return, y_log_return):
    log_regressor = x_log_return/y_log_return
    log_regressor[log_regressor == np.Inf] = np.NaN
    log_regressor[log_regressor == -np.Inf] = np.NaN

    arr = np.array([log_regressor])
    mask = np.isnan(arr)
    idx = np.where(~mask, np.arange(mask.shape[1]), 0)
    arr[mask] = arr[np.nonzero(mask)[0], idx[mask]]
    log_regressor = arr[0]
    return log_regressor

def ecm_result(alpha, residuals):
    return alpha*residuals

def expanding_rolling_mean(signal_np_array, rolling_window=5):
    signal_np_array = np.nan_to_num(signal_np_array)
    cumsum, moving_aves = [0], []
    for i, x in enumerate(signal_np_array, 1):
        cumsum.append(cumsum[i - 1] + x)
        if i > rolling_window:
            moving_av = (cumsum[i] - cumsum[i - rolling_window]) / rolling_window
            moving_aves.append(moving_av)
    moving_aves = np.concatenate([np.array([np.NaN]*rolling_window),moving_aves])
    return moving_aves

def expanding_rolling_zscore(signal_np_array=None, skipping_point = 5, signal_continuum_threshold = 1):
    signal_np_array = np.nan_to_num(signal_np_array)
    try:
        assert skipping_point < len(signal_np_array)
    except:
        log.error('Number of skipping point should be lower than the len of the array')
    zscore = signal_utility.expanding_zscore(signal_np_array, skipping_point=skipping_point,signal_continuum_threshold=1)
    return zscore

def zscore_threshold(my_array,threshold_type='max'):
    my_array = my_array[:-1]
    my_array = np.nan_to_num(my_array)
    if threshold_type == 'max':
        return max(my_array,key=abs)
    elif threshold_type == 'mean':
        return np.mean(my_array)

def check_shift(close, shift):
    if shift == 0:
        return close
    else:
        return close[:-shift]


#!/usr/bin/env python3
# coding: utf-8

""" Tools to display graphs. """

# Built-in packages

# Third party packages
import matplotlib.pyplot as plt

# Local packages

plt.style.use('seaborn')

__all__ = ['DynaGraph']


class DynaGraph:
    """ Object to plot dynamicaly a graph. """

    def __init__(self, title, ax=None):
        plt.ion()

        if not ax:
            self.f, self.ax = plt.subplots(1, 1, figsize=(16, 16))
        else:
            self.ax = ax

        self.ax.set_title(title)

    def plot_df(self, df, **kwargs):
        """ Plot data frome `pd.DataFrame`. """
        self.ax.clear()
        df.plot(ax=self.ax, **kwargs)
        plt.draw()



__author__ = "hugo.inzirillo"

from dataclasses import dataclass
from typing import Dict, List

from pandas import Timestamp

from napoleontoolbox.models.product import Product


@dataclass
class RequestPortfolioAnalysis:
    amount: float
    leverage: float
    composition: Dict[str, float]
    start_date: Timestamp
    end_date: Timestamp
    products: List[Product]



__author__ = "hugo.inzirillo"

import matplotlib.pyplot as plt
import scipy.stats as scs
import statsmodels.api as sm
import statsmodels.tsa.api as smt
from pandas import Series


def tsplot(y, lags=None, figsize=(10, 8), style='bmh'):
    if not isinstance(y, Series):
        y = Series(y)
    with plt.style.context(style):
        fig = plt.figure(figsize=figsize)
        layout = (3, 2)
        ts_ax = plt.subplot2grid(layout, (0, 0), colspan=2)
        acf_ax = plt.subplot2grid(layout, (1, 0))
        pacf_ax = plt.subplot2grid(layout, (1, 1))
        qq_ax = plt.subplot2grid(layout, (2, 0))
        pp_ax = plt.subplot2grid(layout, (2, 1))

        y.plot(ax=ts_ax)
        ts_ax.set_title('Time Series Analysis Plots')
        smt.graphics.plot_acf(y, lags=lags, ax=acf_ax, alpha=0.05)
        smt.graphics.plot_pacf(y, lags=lags, ax=pacf_ax, alpha=0.05)
        sm.qqplot(y, line='s', ax=qq_ax)
        qq_ax.set_title('QQ Plot')
        scs.probplot(y, sparams=(y.mean(), y.std()), plot=pp_ax)

        plt.tight_layout()
        plt.show()



__author__ = "hugo.inzirillo"

from dataclasses import dataclass
from typing import List, Dict

from pandas import Timestamp


@dataclass(frozen=False)
class SignalComputationFilter:
    """
    Position Service
    Mapping of Java Objects
    """
    productCodes: List[str]
    underlyingCodes: List[str] = None
    minTs: str = None
    maxTs: str = None
    lastOnly: bool = False

    def to_dict(self):
        return self.__dict__

    def __str__(self):
        return "SignalComputationFilter{productCodes=" + str(
            self.productCodes) + ", minTs=" + self.minTs + ", maxTs=" + self.maxTs + ", lastOnly=" + str(
            self.lastOnly) + '}'


@dataclass(frozen=True)
class EodQuoteFilter:
    """
    Product Service
    Mapping of Java Objects
    """
    productCodes: List[str]
    sources: List[str] =None
    minDate: str = None
    maxDate: str = None
    lastOnly: bool = False

    def to_dict(self):
        return self.__dict__

    def __str__(self):
        return "EodQuoteFilter{productCodes=" + str(
            self.productCodes) + ", minDate=" + self.minDate + ", maxDate=" + self.maxDate + ", lastOnly=" + str(
            self.lastOnly) + '}'


@dataclass(frozen=True)
class PortFolioFilter:
    compo: Dict[str, float]
    amount: float
    start_date: Timestamp
    end_date: Timestamp



__author__ = "hugo.inzirillo"

from abc import abstractmethod

from napoleontoolbox.napoleon_config_tools.loader.reader import NapoleonConfigReader, Config
from napoleontoolbox.napoleon_config_tools.object import Profile, Env
from napoleontoolbox.napoleon_config_tools.parser.abstractparser import AbstractParserTemplate
from npxlogger import log


class Singleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances.keys():
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class ParserTemplate(AbstractParserTemplate):
    """
    This object is a template of method to implement to create an instance of parser
    """

    @abstractmethod
    def _load(self):
        """
        This method will look up the yaml file attached to your loader
        Returns
        -------

        """
        raise NotImplemented

    @abstractmethod
    def _existing(self):
        """
        This method will check the existence of a config file
        Returns
        -------

        """
        raise NotImplemented

    @abstractmethod
    def _read_file(self):
        """
        This method will read the config file yaml
        Returns
        -------

        """
        raise NotImplemented

    @property
    @abstractmethod
    def config(self) -> Config:
        """
        This property return the Config
        Returns
        -------

        """
        raise NotImplemented

    @property
    @abstractmethod
    def reader(self):
        """
        This property return the Feeder
        Returns
        -------

        """
        raise NotImplemented


class Parser(ParserTemplate):
    """
    This object is the core of config parser
    """

    def __init__(self):
        self.__config = Config()
        self.__reader = NapoleonConfigReader()
        self.__raw_config = dict()

    def _load(self):
        """
        This method will look up the yaml file attached to your loader
        Returns
        -------

        """
        pass

    def _existing(self):
        """
        This method will check the existence of a config file
        Returns
        -------

        """
        pass

    def _read_file(self):
        """
        This method will read the config file yaml
        Returns
        -------

        """
        pass

    @property
    def config(self) -> Config:
        """
        This property return the Config
        Returns
        -------

        """
        return self.__config

    @property
    def reader(self) -> NapoleonConfigReader:
        """
        This property return the Feeder
        Returns
        -------

        """
        return self.__reader

    @property
    def raw_config(self) -> dict:
        """
        This property return the Config
        Returns
        -------

        """
        return self.__raw_config

    @raw_config.setter
    def raw_config(self, _raw_config):
        """
        This property return the Config
        Returns
        -------

        """
        self.__raw_config = _raw_config


@Profile(Env.TEST.value)
class NapoleonToolboxConfigParser(Parser):
    """
    This object will parse the config.yml file to retrieve all the settings
    The object will be automatically triggered
    """
    __metaclass__ = Singleton

    _config = {}

    @property
    def config(self) -> Config:
        return self.__config

    @config.setter
    def config(self, _conf: Config):
        self.__config = _conf

    def parse(self) -> Config:
        """
        This method will parse the config and return a Config object
        Returns
        -------

        """
        self.config = Config()
        _ = {setattr(self.config, key, value) for key, value in self.raw_config.items()}
        log.info("current loader : {config}".format(config=self.config))
        NapoleonToolboxConfigParser._config[self.config.profile] = self.config
        return NapoleonToolboxConfigParser._config[self.config.profile]

    def _read_file(self):
        """
        This method will read the config file yaml
        Returns
        -------

        """

        self.raw_config = self.reader.read()
        return self

    def read(self):
        """
        This method will return the file reader
        Returns
        -------

        """
        return self._read_file()



#!/usr/bin/env python3
# coding: utf-8

""" Object to connect to the NaPoleonX database. """

# Built-in packages
import datetime
import requests
import time
import json

# Third party packages
import pandas as pd
import pickle
from datetime import datetime as dt
#from pandas_datareader import data

from napoleontoolbox.connector import shifting_utility

# Local packages

__all__ = ['NaPoleonXConnector']

def save_pickled_list(list_to_save = None, local_root_directory= '../data/', list_pkl_file_name = 'my_list.pkl'):
    with open(local_root_directory+list_pkl_file_name, 'wb') as f:
        pickle.dump(list_to_save, f)

def load_pickled_list( local_root_directory= '../data/', list_pkl_file_name = 'my_list.pkl'):
    with open(local_root_directory+list_pkl_file_name, 'rb') as f:
        my_list = pickle.load(f)
    return my_list

def request_minute_data_paquet(url, me_ts, ssj):
    r = requests.get(url.format(ssj, me_ts))
    dataframe = None
    try:
        dataframe = pd.DataFrame(json.loads(r.text)['Data']['Data'])
    except Exception as e:
        print('no data')
    return dataframe

def request_hour_data_paquet(url, me_ts, ssj):
    r = requests.get(url.format(ssj, me_ts))
    dataframe = None
    try:
        dataframe = pd.DataFrame(json.loads(r.text)['Data'])
    except Exception as e:
        print('no data')
    return dataframe

def request_day_data_paquet(url, me_ts, ssj):
    r = requests.get(url.format(ssj, me_ts))
    dataframe = None
    try:
        dataframe = pd.DataFrame(json.loads(r.text)['Data'])
    except Exception as e:
        print('no data')
    return dataframe

def save_equity_daily_data(data_df = None, ticker=None, provider = 'napoleon', starting_date=None,running_date=None,local_root_directory=None):
    filename = ticker + '_' + starting_date.strftime('%d_%b_%Y') + '_' + running_date.strftime(
        '%d_%b_%Y') + '_daily_returns.pkl'
    data_df.to_pickle(local_root_directory + filename)

def save_crypto_daily_data(data_df = None, ticker ='crypto_lo', return_pkl_suffix = '_daily_returns.pkl', starting_date=None,running_date=None,local_root_directory=None):
    filename = ticker + '_' + starting_date.strftime('%d_%b_%Y') + '_' + running_date.strftime(
        '%d_%b_%Y') + return_pkl_suffix
    print(f'saving {local_root_directory + filename}')
    data_df.to_pickle(local_root_directory + filename)

def fetch_equity_daily_data(data = None, ticker=None, provider = 'yahoo', starting_date=None,running_date=None,local_root_directory=None):
    yahoo_tickers_equivalents = {
        'SPX' : '^GSPC',
        'SX5E' : '^STOXX50E',
    }

    yahoo_ticker = ticker
    try:
        yahoo_ticker = yahoo_tickers_equivalents[ticker]
    except:
        print('No yahoo ticker equivalent for this ticker. Check dictionary of yahoo tickers equivalents')

    if provider == 'yahoo':
        data_df = None# data.DataReader(yahoo_ticker, provider, starting_date, running_date)
        data_df.columns = map(str.lower, data_df.columns)
        filename = ticker+'_'+starting_date.strftime('%d_%b_%Y')+'_'+running_date.strftime('%d_%b_%Y')+'_daily_returns.pkl'
        data_df.to_pickle(local_root_directory+filename)
        return data_df
    else:
        print('This provider has not been set up yet.')
        return

def fetch_delayed_crypto_daily_data(frequence = 'daily', ssj=None, local_root_directory=None, daily_return_pkl_filename_suffix='_daily_returns.pkl', refetch_all=True,daily_crypto_starting_day='2012-01-01', daily_crypto_ending_day=None,ssj_against='USDT', exchange = 'binance', delay = 0):
    assert delay>0
    hourly_crypto_df = fetch_crypto_hourly_data(ssj=ssj,local_root_directory=local_root_directory,
                                                                   refetch_all=refetch_all,
                                                                   daily_crypto_starting_day=daily_crypto_starting_day,
                                                                   daily_crypto_ending_day=daily_crypto_ending_day)

    daily_crypto_df = shifting_utility.shift_daily_crypto_compare_signal(hourly_data = hourly_crypto_df, ssj=ssj, local_root_directory=local_root_directory,
                                                                     shift=delay, starting_date=daily_crypto_starting_day,
                                                                     running_date=daily_crypto_ending_day, frequence='daily')

    freqly_pkl_file_name_suffix = f'_{frequence}_returns.pkl'
    dates_stub = daily_crypto_starting_day.strftime('%d_%b_%Y') + '_' + daily_crypto_ending_day.strftime('%d_%b_%Y')
    saving_return_path = f'{local_root_directory}{ssj}_{dates_stub}_{delay}{freqly_pkl_file_name_suffix}'
    daily_crypto_df.to_pickle(saving_return_path)
    return daily_crypto_df



def fetch_crypto_daily_data(ssj=None, local_root_directory=None, daily_return_pkl_filename_suffix='_daily_returns.pkl', refetch_all=True,daily_crypto_starting_day='2012-01-01', daily_crypto_ending_day=None,ssj_against='USDT', exchange = 'binance'):
    dates_stub = daily_crypto_starting_day.strftime('%d_%b_%Y') + '_' + daily_crypto_ending_day.strftime('%d_%b_%Y')
    pickle_saving_path = local_root_directory+ssj+'_'+dates_stub+daily_return_pkl_filename_suffix
    if refetch_all:
        year = datetime.datetime.now().year
        month = datetime.datetime.now().month
        day = datetime.datetime.now().day
        hour = datetime.datetime.utcnow().hour
        ts = datetime.datetime(year, month, day, tzinfo=datetime.timezone.utc).timestamp() + hour * 3600
        ts1 = ts - 2001 * 3600 * 24
        ts2 = ts1 - 2001 * 3600 * 24
        ts3 = ts2 - 2001 * 3600 * 24
        ts4 = ts3 - 2001 * 3600 * 24
        ts5 = ts4 - 2001 * 3600 * 24
        ts6 = ts5 - 2001 * 3600 * 24
        ts7 = ts6 - 2001 * 3600 * 24
        ts8 = ts7 - 2001 * 3600 * 24

        print('Loading data')
        day_url_request = 'https://min-api.cryptocompare.com/data/histoday?fsym={}&tsym='+ssj_against+'&e='+exchange+'&toTs={}&limit=2000'
        dataframe = None
        for me_timestamp in [ts8,ts7,ts6,ts5,ts4,ts3,ts2,ts1,ts]:
            print('waiting')
            df = request_day_data_paquet(day_url_request, me_timestamp, ssj)
            if df is not None:
                if dataframe is not None:
                    dataframe = dataframe.append(df, ignore_index=True)
                else:
                    dataframe = df.copy()


        dataframe['time'] = pd.to_datetime(dataframe['time'], unit='s')
        dataframe = dataframe.sort_values(by=['time'])
        dataframe = dataframe.rename(columns={"time": "date"}, errors="raise")
        dataframe = dataframe.set_index(dataframe['date'])
        dataframe = dataframe.drop(columns=['date'])
        print('size fetched')
        print(dataframe.shape)
        dataframe = dataframe[dataframe.index >= daily_crypto_starting_day]
        dataframe = dataframe[dataframe.index <= daily_crypto_ending_day]
        print('size filtered after '+str(daily_crypto_starting_day))
        print(dataframe.shape)
        print(f'writing {pickle_saving_path}')
        dataframe.to_pickle(pickle_saving_path)
    else:
        print(f'reading {pickle_saving_path}')
        dataframe = pd.read_pickle(pickle_saving_path)
    return dataframe


def fetch_rest_crypto_compare_last_minute_ohlc(ssj, against, limit = 1):
    last_minute_url = f'https://min-api.cryptocompare.com/data/v2/histominute?fsym={ssj}&tsym={against}&limit={limit}'
    r = requests.get(last_minute_url)
    dataframe = None
    try:
        dataframe = pd.DataFrame(json.loads(r.text)['Data']['Data'])
    except Exception as e:
        print('no data')

    dataframe['time'] = pd.to_datetime(dataframe['time'], unit='s')
    dataframe = dataframe.sort_values(by=['time'])
    dataframe = dataframe.rename(columns={"time": "date"}, errors="raise")
    dataframe = dataframe.set_index(dataframe['date'])
    dataframe = dataframe.drop(columns=['date'])

    return dataframe




def fetch_crypto_minutely_data(ssj=None, local_root_directory=None, minutely_return_pkl_filename_suffix=None, running_date= None, daily_crypto_starting_day='2012-01-01', daily_crypto_ending_day=None, refetch_all=True, ssj_against='USDT', exchange = 'binance'):
    dates_stub = daily_crypto_starting_day.strftime('%d_%b_%Y') + '_' + daily_crypto_ending_day.strftime('%d_%b_%Y')
    pickle_saving_path = local_root_directory+ssj+'_'+dates_stub+minutely_return_pkl_filename_suffix
    if refetch_all:
        year = datetime.datetime.now().year
        month = datetime.datetime.now().month
        day = datetime.datetime.now().day
        hour = datetime.datetime.utcnow().hour
        ts = datetime.datetime(year, month, day, tzinfo=datetime.timezone.utc).timestamp() + hour * 3600
        ts1 = ts - 2001 * 60
        ts2 = ts1 - 2001 * 60
        ts3 = ts2 - 2001 * 60
        ts4 = ts3 - 2001 * 60
        ts5 = ts4 - 2001 * 60
        ts6 = ts5 - 2001 * 60
        ts7 = ts6 - 2001 * 60
        ts8 = ts7 - 2001 * 60
        ts9 = ts8 - 2001 * 60
        ts10 = ts9 - 2001 * 60
        ts11 = ts10 - 2001 * 60
        ts12 = ts11 - 2001 * 60
        print('Loading data')
        binance_min_usdt_url = 'https://min-api.cryptocompare.com/data/v2/histominute?fsym={}&tsym='+ssj_against+'&e='+exchange+'&toTs={}&limit=2000'

        dataframe = None
        for me_timestamp in [ts12,ts11,ts10,ts9,ts8,ts7,ts6,ts5,ts4,ts3,ts2,ts1,ts]:
            print('waiting')
            df = request_minute_data_paquet(binance_min_usdt_url, me_timestamp, ssj)
            if df is not None:
                if dataframe is not None:
                    dataframe = dataframe.append(df, ignore_index=True)
                else:
                    dataframe = df.copy()

        dataframe['time'] = pd.to_datetime(dataframe['time'], unit='s')
        dataframe = dataframe.sort_values(by=['time'])
        dataframe = dataframe.rename(columns={"time": "date"}, errors="raise")
        dataframe = dataframe.set_index(dataframe['date'])
        dataframe = dataframe.drop(columns=['date'])
        print('size fetched')
        print(dataframe.shape)
        dataframe.to_pickle(pickle_saving_path)
    else:
        dataframe = pd.read_pickle(pickle_saving_path)
    return dataframe

def fetch_local_crypto_hourly_data(ssj=None, local_root_directory=None, hourly_return_pkl_filename_suffix='_hourly_returns.pkl',refetch_all=True, daily_crypto_starting_day='2012-01-01', daily_crypto_ending_day=None, ssj_against='USDT', exchange = 'binance',local_path= f'/Users/stefanduprey/Documents/My_StephaneStrategy/'):
    dates_stub = daily_crypto_starting_day.strftime('%d_%b_%Y') + '_' + daily_crypto_ending_day.strftime('%d_%b_%Y')

    pickle_saving_path = local_root_directory + ssj + '_' + dates_stub + hourly_return_pkl_filename_suffix

    if ssj == 'BTC':
        dataframe = pd.read_pickle(local_path + f'btc_hourly.pkl')

    elif ssj == 'ETH':
        dataframe = pd.read_pickle(local_path + f'eth_hourly.pkl')

    dataframe.columns =  ['date','close']

    dataframe['date'] = pd.to_datetime(dataframe['date'])
    dataframe = dataframe.set_index(dataframe['date'])
    dataframe = dataframe.drop(columns=['date'])
    dataframe = dataframe.sort_index()
    dataframe['open'] = dataframe['close'].shift(1)

    print('size fetched')
    print(dataframe.shape)
    dataframe = dataframe[dataframe.index >= daily_crypto_starting_day]
    print('size filtered after '+str(daily_crypto_starting_day))
    print(dataframe.shape)
    dataframe.to_pickle(pickle_saving_path)
    return dataframe


def fetch_crypto_hourly_data(ssj=None, local_root_directory=None, hourly_return_pkl_filename_suffix='_hourly_returns.pkl',refetch_all=True, daily_crypto_starting_day='2012-01-01', daily_crypto_ending_day=None, ssj_against='USDT', exchange = 'binance'):
    dates_stub = daily_crypto_starting_day.strftime('%d_%b_%Y') + '_' + daily_crypto_ending_day.strftime('%d_%b_%Y')
    pickle_saving_path = local_root_directory + ssj + '_' + dates_stub + hourly_return_pkl_filename_suffix
    if refetch_all:
        year = datetime.datetime.now().year
        month = datetime.datetime.now().month
        day = datetime.datetime.now().day
        hour = datetime.datetime.utcnow().hour
        ts = datetime.datetime(year, month, day, tzinfo=datetime.timezone.utc).timestamp() + hour * 3600
        ts1 = ts - 2001 * 3600
        ts2 = ts1 - 2001 * 3600
        ts3 = ts2 - 2001 * 3600
        ts4 = ts3 - 2001 * 3600
        ts5 = ts4 - 2001 * 3600
        ts6 = ts5 - 2001 * 3600
        ts7 = ts6 - 2001 * 3600
        ts8 = ts7 - 2001 * 3600
        ts9 = ts8 - 2001 * 3600
        ts10 = ts9 - 2001 * 3600
        ts11 = ts10 - 2001 * 3600
        ts12 = ts11 - 2001 * 3600
        ts13 = ts12 - 2001 * 3600
        ts14 = ts13 - 2001 * 3600
        ts15 = ts14 - 2001 * 3600
        ts16 = ts15 - 2001 * 3600
        ts17 = ts16 - 2001 * 3600
        ts18 = ts17 - 2001 * 3600

        print('Loading data')
        hours_url_request = 'https://min-api.cryptocompare.com/data/histohour?fsym={}&tsym=' + ssj_against + '&e=' + exchange + '&toTs={}&limit=2000'
        dataframe = None
        for me_timestamp in [ts18, ts17, ts16, ts15, ts14, ts13, ts12, ts11, ts10, ts9, ts8, ts7, ts6, ts5, ts4, ts3, ts2, ts1, ts]:
            print('waiting')
            df = request_hour_data_paquet(hours_url_request, me_timestamp, ssj)
            if df is not None:
                if dataframe is not None:
                    dataframe = dataframe.append(df, ignore_index=True)
                else:
                    dataframe = df.copy()

        dataframe['time'] = pd.to_datetime(dataframe['time'], unit='s')
        dataframe = dataframe.sort_values(by=['time'])
        dataframe = dataframe.rename(columns={"time": "date"}, errors="raise")
        dataframe = dataframe.set_index(dataframe['date'])
        dataframe = dataframe.drop(columns=['date'])
        print('size fetched')
        print(dataframe.shape)
        dataframe = dataframe[dataframe.index >= daily_crypto_starting_day]
        print('size filtered after '+str(daily_crypto_starting_day))
        print(dataframe.shape)
        dataframe.to_pickle(pickle_saving_path)
    else:
        dataframe = pd.read_pickle(pickle_saving_path)
    return dataframe

def fetch_crypto_tuple_hourly_data(ssj=None, pickle_saving_path=None, refetch_all=True, frequency=4):
    if refetch_all:
        year = datetime.datetime.now().year
        month = datetime.datetime.now().month
        day = datetime.datetime.now().day
        hour = datetime.datetime.utcnow().hour
        ts = datetime.datetime(year, month, day, tzinfo=datetime.timezone.utc).timestamp() + hour * 3600
        ts1 = ts - 2001 * 3600
        ts2 = ts1 - 2001 * 3600
        ts3 = ts2 - 2001 * 3600
        ts4 = ts3 - 2001 * 3600
        ts5 = ts4 - 2001 * 3600
        ts6 = ts5 - 2001 * 3600
        ts7 = ts6 - 2001 * 3600
        ts8 = ts7 - 2001 * 3600
        ts9 = ts8 - 2001 * 3600
        ts10 = ts9 - 2001 * 3600
        ts11 = ts10 - 2001 * 3600
        ts12 = ts11 - 2001 * 3600
        ts13 = ts12 - 2001 * 3600
        print('Loading data')
        r = requests.get(
            'https://min-api.cryptocompare.com/data/histohour?fsym={}&tsym=USD&toTs={}&limit=2000'.format(ssj, ts13))
        dataframe = pd.DataFrame(json.loads(r.text)['Data'])
        print('waiting')
        time.sleep(1)
        r = requests.get(
            'https://min-api.cryptocompare.com/data/histohour?fsym={}&tsym=USD&toTs={}&limit=2000'.format(ssj, ts12))
        df = pd.DataFrame(json.loads(r.text)['Data'])
        dataframe = dataframe.append(df, ignore_index=True)
        print('waiting')
        time.sleep(1)
        r = requests.get(
            'https://min-api.cryptocompare.com/data/histohour?fsym={}&tsym=USD&toTs={}&limit=2000'.format(ssj, ts11))
        df = pd.DataFrame(json.loads(r.text)['Data'])
        dataframe = dataframe.append(df, ignore_index=True)
        print('waiting')
        time.sleep(1)
        r = requests.get(
            'https://min-api.cryptocompare.com/data/histohour?fsym={}&tsym=USD&toTs={}&limit=2000'.format(ssj, ts10))
        df = pd.DataFrame(json.loads(r.text)['Data'])
        dataframe = dataframe.append(df, ignore_index=True)
        print('waiting')
        time.sleep(1)
        r = requests.get(
            'https://min-api.cryptocompare.com/data/histohour?fsym={}&tsym=USD&toTs={}&limit=2000'.format(ssj, ts9))
        df = pd.DataFrame(json.loads(r.text)['Data'])
        dataframe = dataframe.append(df, ignore_index=True)
        print('waiting')
        time.sleep(1)
        r = requests.get(
            'https://min-api.cryptocompare.com/data/histohour?fsym={}&tsym=USD&toTs={}&limit=2000'.format(ssj, ts8))
        df = pd.DataFrame(json.loads(r.text)['Data'])
        dataframe = dataframe.append(df, ignore_index=True)
        print('waiting')
        time.sleep(1)
        r = requests.get(
            'https://min-api.cryptocompare.com/data/histohour?fsym={}&tsym=USD&toTs={}&limit=2000'.format(ssj, ts7))
        df = pd.DataFrame(json.loads(r.text)['Data'])
        dataframe = dataframe.append(df, ignore_index=True)
        print('waiting')
        time.sleep(1)
        r = requests.get(
            'https://min-api.cryptocompare.com/data/histohour?fsym={}&tsym=USD&toTs={}&limit=2000'.format(ssj, ts6))
        df = pd.DataFrame(json.loads(r.text)['Data'])
        dataframe = dataframe.append(df, ignore_index=True)
        print('waiting')
        time.sleep(1)
        r = requests.get(
            'https://min-api.cryptocompare.com/data/histohour?fsym={}&tsym=USD&toTs={}&limit=2000'.format(ssj, ts5))
        df = pd.DataFrame(json.loads(r.text)['Data'])
        dataframe = dataframe.append(df, ignore_index=True)
        time.sleep(1)
        r = requests.get(
            'https://min-api.cryptocompare.com/data/histohour?fsym={}&tsym=USD&toTs={}&limit=2000'.format(ssj, ts4))
        df = pd.DataFrame(json.loads(r.text)['Data'])
        dataframe = dataframe.append(df, ignore_index=True)
        r = requests.get(
            'https://min-api.cryptocompare.com/data/histohour?fsym={}&tsym=USD&toTs={}&limit=2000'.format(ssj, ts3))
        df = pd.DataFrame(json.loads(r.text)['Data'])
        dataframe = dataframe.append(df, ignore_index=True)
        r = requests.get(
            'https://min-api.cryptocompare.com/data/histohour?fsym={}&tsym=USD&toTs={}&limit=2000'.format(ssj, ts2))
        df = pd.DataFrame(json.loads(r.text)['Data'])
        dataframe = dataframe.append(df, ignore_index=True)
        r = requests.get(
            'https://min-api.cryptocompare.com/data/histohour?fsym={}&tsym=USD&toTs={}&limit=2000'.format(ssj, ts1))
        df = pd.DataFrame(json.loads(r.text)['Data'])
        dataframe = dataframe.append(df, ignore_index=True)
        r = requests.get(
            'https://min-api.cryptocompare.com/data/histohour?fsym={}&tsym=USD&toTs={}&limit=2000'.format(ssj, ts))
        df = pd.DataFrame(json.loads(r.text)['Data'])
        dataframe = dataframe.append(df, ignore_index=True)

        dataframe['time'] = pd.to_datetime(dataframe['time'], unit='s')
        dataframe = dataframe.sort_values(by=['time'])
        dataframe = dataframe.rename(columns={"time": "date"}, errors="raise")
        dataframe = dataframe.set_index(dataframe['date'])
        dataframe = dataframe.drop(columns=['date'])
        dataframe = dataframe.iloc[::frequency, :]
        print('size fetched')
        print(dataframe.shape)
        dataframe.to_pickle(pickle_saving_path)
    else:
        dataframe = pd.read_pickle(pickle_saving_path)
    return dataframe



class NaPoleonXConnector(object):
    """ Object to connect to the NaPoleonX database.

    Parameters
    ----------
    username, password, client_id, client_secret : str
        Identifier to request the API.

    Attributes
    ----------
    token : str
        token to identify client.

    Methods
    -------
    get_data
    get_dataframe

    """

#    url_auth = 'https://api.napoleonx.ai/auth-service/oauth/token'
#    url = 'https://api.napoleonx.ai/quote-service/v1/eod-quote/filter'
#    signal_url = 'https://api.napoleonx.ai/signal-service/v1/signal-computation/filter'
#    url_product = 'https://api.napoleonx.ai/product-service/v1/product/'

#    url_auth = 'http://15.237.36.185:9000/auth-service/oauth/token'

    url_auth = 'http://15.237.36.185:9000/oauth/token'
    url = 'http://15.237.36.185:9000/v1/quote-service/filter'
    signal_url = 'http://15.237.36.185:9000/v1/position-service/filter'
    url_product = 'http://15.237.36.185:9000/v1/product-service/'

    # url_auth = 'https://api.napoleonx.ai/oauth/token'
    # url = 'https://api.napoleonx.ai/v1/quote-service/filter'
    # signal_url = 'https://api.napoleonx.ai/v1/position-service/filter'
    # url_product = 'https://api.napoleonx.ai/v1/product-service/'

    #    url_auth = 'http://15.237.36.185:9000/auth-service/oauth/token'

    data = {
        'grant_type': 'password',
#        'scope': 'read write',
        'scope': 'read'
    }

    headers = {
        "Accept": "application/json",
        "Content-Type": "application/x-www-form-urlencoded;charset=UTF-8",
    }


    def __init__(self, username, password, client_id, client_secret):
        """ Initialize object. """
        auth = requests.auth.HTTPBasicAuth(client_id, client_secret)

        self.data.update({'username': username, 'password': password})

        response_w = requests.post(
            self.url_auth,
            data=self.data,
            headers=self.headers,
            auth=auth
        )
        response=response_w.json()

        # Set token
        self.token = response['access_token']
        self.token_type = response['token_type']
        self._set_header()

    def _set_header(self):
        self.headers = {
            'Content-Type': 'application/json',
            'Accept': 'application/json',
            'Authorization': "{} {}".format(
                self.token_type,
                self.token
            ),
            'cache-control': "no-cache"
        }

    def _post_request(self, url, **kwargs):
        data = json.dumps(kwargs)
        return requests.post(url, data=data, headers=self.headers).json()

    def _get_request(self, url, **kwargs):
        data = json.dumps(kwargs)
        json_data = requests.get(url, data=data, headers=self.headers).json()
        return json_data

    def get_signal_data(self, productCodes, minTs, maxTs, lastOnly):
        return self._post_request(
            self.signal_url,
            productCodes=productCodes,
            minTs=minTs,
            maxTs=maxTs,
            lastOnly=lastOnly
        )

    def get_data(self, productCodes, minDate, maxDate):
        """ Request data.

        Parameters
        ----------
        productCodes : list of str
            List of codes of assets.
        minDate, maxDate : str
            Respectively the first and last date of the data.
        kwargs
            Key word arguments for requests the database.

        Returns
        -------
        dict
            Data.

        """
        return self._post_request(
            self.url,
            productCodes=productCodes,
            minDate=minDate,
            maxDate=maxDate,
            lastOnly = False
        )

    def get_product_data(self, productCodes):
        return self._post_request(
            self.url_product,
            productCodes=productCodes
        )

    def _set_dataframe(self, data, assets, keep=['close']):
        """ Clean, sort and set data into the dataframe.

        Parameters
        ----------
        data : pandas.DataFrame
            Dataframe of not ordered data.
        assets : list of str
            List of code of assets.
        keep : list of str
            List of columns to keep into the dataframe.

        Returns
        -------
        pandas.DataFrame
            Where each column is an asset price and each row is a date.

        """
        columns, asset_col = {}, []
        for asset in assets:
            if len(keep) > 1:
                columns[asset] = {k: k + '_' + asset for k in keep}
                asset_col += [k + '_' + asset for k in keep]

            else:
                columns[asset] = {keep[0]: asset}
                asset_col += [asset]

        df = pd.DataFrame(
            columns=asset_col,
            index=sorted(data.date.drop_duplicates()),
        )

        for asset in assets:
            sub_df = (data.loc[data.productCode == asset, keep + ['date']]
                          .set_index('date')
                          .sort_index()
                          .rename(columns=columns[asset]))
            asset_col = list(columns[asset].values())
            df.loc[sub_df.index, asset_col] = sub_df.loc[:, asset_col]
        return df


    def get_vm_rates_dataframe_from_currency(self, currencies, minDate, maxDate):
        vm_rates_product_codes = [f'VM-EUR-{currency}' for currency in currencies if currency != 'EUR']
        return self.get_dataframe(vm_rates_product_codes, minDate, maxDate)


    def get_vm_rates_dataframe(self, productCodes, minDate, maxDate):
        currencies = set([self.get_currency(product_code) for product_code in productCodes])
        vm_rates_product_codes = [f'VM-EUR-{currency}' for currency in currencies if currency != 'EUR']
        return self.get_dataframe(vm_rates_product_codes, minDate, maxDate)

    def get_dataframe(self, productCodes, minDate, maxDate, keep=['close'],
                      process_na=None):
        """ Request data, clean, sort, and set into pandas dataframe.

        Parameters
        ----------
        productCodes : list of str
            List of codes of assets.
        minDate, maxDate : str
            Respectively the first and last date of the data.
        keep : list of str
            List of columns to keep into the dataframe.
        process_na : {None, 'fill', 'drop'}
            - If None don't return dataframe with nan values.
            - If 'fill' replace nan by the last observation or by the next one.
            - If 'drop' drop the nan values.

        Returns
        -------
        pd.DataFrame
            Cleaned dataframe.

        """
        json_data = self.get_data(productCodes, minDate, maxDate)
        data = pd.DataFrame(json_data['data'])
        df = self._set_dataframe(data, assets=productCodes, keep=keep)
        df.index = pd.to_datetime(df.index, format='%Y-%m-%d')

        if process_na == 'fill':

            return df.fillna(method='ffill').fillna(method='bfill')

        elif process_na == 'drop':

            return df.dropna()

        return df

    def get_signal_dataframe(self, productCodes, minDate, maxDate, last_only = True):
        minTs = dt.strptime(minDate, '%Y-%m-%d').strftime('%Y-%m-%d') + 'T00:00:00'
        maxTs = dt.strptime(maxDate, '%Y-%m-%d').strftime('%Y-%m-%d') + 'T23:59:59'
        signals_df = pd.DataFrame(self.get_signal_data(productCodes,  minTs, maxTs, last_only)['data'])
        return signals_df

    def get_underlyings(self, productCode):
        data = self._get_request(self.url_product + productCode)['data']
        return data['underlyings']

    def get_currency(self, productCode):
        data = self._get_request(self.url_product + productCode)
        return data['data']['currency']

